#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cmdline.h"
#include "geodesie.h"
#include "misc.h"
#include "raf98.h"

#include "main.h"

#define _VERSION "0.2 beta"
#define _DATE    "Mar 2nd, 2005"

// --- Options flags -----------------------------------------------------------
#define HELP          0 // Print the help
#define ANTENNA       1 // Heigth value of the antenna
#define INPUT_FILE    2 // Read input in <input-file>
#define INPUT_SYSG    3 // Input geodetic system
#define INPUT_PROJ    4 // Input projection
#define INPUT_COORD   5 // Input coordinates type
#define OUTPUT_FILE   6 // Write output in <output-file>
#define OUTPUT_SYSG   7 // Output geodetic system
#define OUTPUT_PROJ   8 // Output projection type
#define OUTPUT_COORD  9 // Output coordinates type
#define VERBOSE      10 // Verbose mode ON
#define VERSION      11 // Show software version

int flags[10]={false};

#define CARTESIENNES  0
#define GEOGRAPHIQUES 1
#define PLANES        2

// --- Functions ---------------------------------------------------------------

void help(void) {
   printf("Usage: convert -i<input-file> [options]\n"
          "Options:\n"
          "  -h --help              Print this help\n"
          "  -ha                    Antenna GPS height value\n"
          "  -i<input-file>         Read input in <input-file>\n"
          "    -is=[system]         Input geodetic system\n"
          "    -ip=                 Input projection\n"
          "    -ic=g                Input coordinates type (geographic)\n"
          "  -o<output-file>        Write output in <output-file>\n"
          "    -os=[system]         Output geodetic system\n"
          "    -op=[projection]     Output projection type\n"
          "    -oc=p                Output coordinates type (flat)\n"
          "  -v --verbose           Verbose mode ON\n"
          "  -V --version           Show software version\n"
          "\n"
          "  Direction  Systems  Projections\n"
          "  input      WGS84    -none-\n"
          "  output     NTF      Lb1      Lambert I (Nord)\n"
          "  output     NTF      Lb2      Lambert II (center)\n"
          "  output     NTF      Lb2e     Lambert II extended\n"
          "  output     NTF      Lb3      Lambert III (Sud)\n"
          "  output     NTF      Lb4      Lambert IV (Corse)\n"/*
          "  output     RGF93    Lb93     Lambert 93 (grand champ)\n"*/);
}

void ignore(const char *str) {
   fprintf(stderr, "MIS-convert: [ATTENTION] Ignored parameter: %s\n", str);
}

void version(void) {
   printf("MIS-convert -- by BadFox <url: http://badfox.mine.nu/>\n");
   printf("version %s [%s]\n", _VERSION, _DATE);
   printf("Magic Instinct Software <url: http://www.justmagic.com/>\n");
}

void wrong(const char *str) {
   fprintf(stderr, "MIS-convert: [ERROR] Wrong parameter: %s\n", str);
}

// --- Main --------------------------------------------------------------------
int main(int argc, char *argv[]) {
   // The max filename length under windows: 256 characters on NTFS as well as VFAT and FAT32.
   char inputFilename[256]="", outputFilename[256]="";
   data in={0}, out={0};
   double ha=0.0;

   // Get the parameters and filenames
   int n=1;

   for(n=1; n<argc; n++) {
      if(argv[n][0] == '-') {
         int len=strlen(argv[n]);

         switch(argv[n][1]) {
            case '-': // --...
            {
               char temp[256]="";
               substr(temp, argv[n], 1, len-1);

               if(strcmp(temp, "-help") == 0) { // --help
                  if(!flags[HELP]) {
                     flags[HELP]=true;
                     help();

                     return 0;
                  } else {
                     ignore(argv[n]);
                  }
               } else if(strcmp(temp, "-verbose") == 0) { // --verbose
                  if(!flags[VERBOSE]) {
                     flags[VERBOSE]=true;
                  } else {
                     ignore(argv[n]);
                  }
               } else if(strcmp(temp, "-version") == 0) { // --version
                  if(!flags[VERSION]) {
                     flags[VERSION]=true;
                     version();

                     return 0;
                  } else {
                     ignore(argv[n]);
                  }
               } else {
                  wrong(argv[n]);
               }
            }
            break;

            case 'h':
            {
               if(len == 2) { // -h
                  if(!flags[HELP]) {
                     flags[HELP]=true;
                     help();

                     return 0;
                  } else {
                     ignore(argv[n]);
                  }
               } else if((len > 4) && (argv[n][3] == '=')) { // -ha=[value]
                  if(!flags[ANTENNA]) {
                     flags[ANTENNA]=true;

                     char temp[256]="";
                     substr(temp, argv[n], 4, len-4);

                     ha=atof(temp);
                  } else {
                     ignore(argv[n]);
                  }
               } else {
                  wrong(argv[n]);
               }
            }
            break;

            case 'i':
            {
               if((len > 4) && (argv[n][3] == '=')) {
                  switch(argv[n][2]) {
                     case 's': // -is=[system]
                     {
                        char temp[256]="";
                        substr(temp, argv[n], 4, len-4);

                        strtolower(temp, temp);

                        if(strcmp(temp, "wgs84") == 0) { // -is=WGS84
                           flags[INPUT_SYSG]=true;
                           in.sysg=0;

                           in.cstes.a=6378137.0;
                           in.cstes.f=1/298.257223563;

                           in.cstes.e=E(in.cstes.a, in.cstes.a*(1.0-in.cstes.f)); // e=f(a, b) w/ b=a(1-f)
                        } else {
                           wrong(argv[n]);
                        }
                     }
                     break;

                     case 'p': // -ip=[projection]
                     {
                        if(!flags[INPUT_PROJ]) {
                           flags[INPUT_PROJ]=true;
                        } else {
                           ignore(argv[n]);
                        }
                     }
                     break;

                     case 'c': // -ic=[type]
                     {
                        if(!flags[INPUT_COORD]) {
                           flags[INPUT_COORD]=true;
                        } else {
                           ignore(argv[n]);
                        }
                     }
                     break;

                     default:
                     {
                        wrong(argv[n]);
                     }
                     break;
                  }
               } else if(len > 2) { //-i<input-file>
                  if(!flags[INPUT_FILE]) {
                     flags[INPUT_FILE]=true;

                     char filename[256]="";
                     FILE *pTemp=NULL;

                     substr(filename, argv[n], 2, len-2);
                     sprintf(inputFilename, "%s", filename);

                     if((pTemp=fopen(filename, "rt")) == NULL) {
                        fprintf(stderr, "MIS-convert: [ERROR] File `%s' does not exist.\n", filename);
                        return 1;
                     } else {
                        strcpy(inputFilename, filename);
                        fclose(pTemp);
                     }
                  } else {
                     ignore(argv[n]);
                  }
               } else {
                  wrong(argv[n]);
               }
            }
            break;

            case 'o':
            {
               if((len > 4) && (argv[n][3] == '=')) {
                  switch(argv[n][2]) {
                     case 's': // -os=[system]
                     {
                        if(!flags[OUTPUT_SYSG]) {
                           char temp[256]="";
                           substr(temp, argv[n], 4, len-4);

                           strtolower(temp, temp);

                           if(strcmp(temp, "ntf") == 0) { // -os=NTF
                              flags[OUTPUT_SYSG]=true;
                              out.sysg=1;

                              // Ellipsoide Clarke 1880 IGN
                              out.cstes.a=6378249.2;
                              out.cstes.b=6356515.0;

                              out.cstes.e=E(out.cstes.a, out.cstes.b);
                              out.cstes.lambda0=deg2rad(2.337229167); // 0 grades Paris (= 2 20' 14,025" E Greenwich)

                              out.cstes.T.x=-168.0;
                              out.cstes.T.y=-60.0;
                              out.cstes.T.z=320.0;
                              out.cstes.D=0.0;
                              out.cstes.R.x=0.0;
                              out.cstes.R.y=0.0;
                              out.cstes.R.z=0.0;
                           } else if(strcmp(temp, "rgf93") == 0) { // -os=RGF93
                              flags[OUTPUT_SYSG]=true;
                              out.sysg=2;

                              // Ellipsoide IAG GRS 80
                              out.cstes.a=6378137.0;
                              out.cstes.f=1/298.257222101;

                              out.cstes.e=E(in.cstes.a, in.cstes.a*(1.0-in.cstes.f)); // e=f(a, b) w/ b=a(1-f)
                              out.cstes.lambda0=deg2rad(3.0); // ? grades Paris (= 3 E Greenwich)
   //                           out.cstes.phiv0=deg2rad(?); // (= 46 30' N Greenwich)

                              out.cstes.T.x=-168.0;
                              out.cstes.T.y=-60.0;
                              out.cstes.T.z=320.0;
                              out.cstes.D=0.0;
                              out.cstes.R.x=0.0;
                              out.cstes.R.y=0.0;
                              out.cstes.R.z=0.0;
                           } else {
                              wrong(argv[n]);
                           }
                        } else {
                           ignore(argv[n]);
                        }
                     }
                     break;

                     case 'p': // -op=[projection]
                     {
                        if(!flags[OUTPUT_PROJ]) {

                           char temp[256]="";
                           substr(temp, argv[n], 4, len-4);

                           strtolower(temp, temp);

                           if(strcmp(temp, "lb1") == 0) { // -op=Lb1
                              flags[OUTPUT_PROJ]=true;
                              out.proj=1;

                              out.cstes.n=0.7604059656;
                              out.cstes.C=11603796.98;
                              out.cstes.Xs=600000.0;
                              out.cstes.Ys=5657616.674;
                           } else if(strcmp(temp, "lb2") == 0) { // -op=Lb2
                              flags[OUTPUT_PROJ]=true;
                              out.proj=2;

                              out.cstes.n=0.7289686274;
                              out.cstes.C=11745793.39;
                              out.cstes.Xs=600000.0;
                              out.cstes.Ys=6199695.768;
                           } else if(strcmp(temp, "lb2e") == 0) { // -op=Lb2e
                              flags[OUTPUT_PROJ]=true;
                              out.proj=3;

                              out.cstes.n=0.7289686274;
                              out.cstes.C=11745793.39;
                              out.cstes.Xs=600000.0;
                              out.cstes.Ys=8199695.768;
                           } else if(strcmp(temp, "lb3") == 0) { // -op=Lb3
                              flags[OUTPUT_PROJ]=true;
                              out.proj=4;

                              out.cstes.n=0.6959127966;
                              out.cstes.C=11947992.52;
                              out.cstes.Xs=600000.0;
                              out.cstes.Ys=6791905.085;
                           } else if(strcmp(temp, "lb4") == 0) { // -op=Lb4
                              flags[OUTPUT_PROJ]=true;
                              out.proj=5;

                              out.cstes.n=0.6712679322;
                              out.cstes.C=12136281.99;
                              out.cstes.Xs=234.358;
                              out.cstes.Ys=7239161.542;
                           } else if(strcmp(temp, "lb93") == 0) { // -op=Lb93
                              flags[OUTPUT_PROJ]=true;
                              out.proj=6;

                              out.cstes.n=0.7256077650;
                              out.cstes.C=11754255.426;
                              out.cstes.Xs=700000.000;
                              out.cstes.Ys=12655612.050;
                           } else {
                              wrong(argv[n]);
                           }
                        } else {
                           ignore(argv[n]);
                        }
                     }
                     break;

                     case 'c': // -oc=[type]
                     {
                        if(!flags[OUTPUT_COORD]) {
                           flags[OUTPUT_COORD]=true;
                        } else {
                           ignore(argv[n]);
                        }
                     }
                     break;

                     default:
                     {
                        wrong(argv[n]);
                     }
                     break;
                  }
               } else if(len > 2) { // -o<output-file>
                  if(!flags[OUTPUT_FILE]) {
                     flags[OUTPUT_FILE]=true;

                     char filename[256]="";

                     substr(filename, argv[n], 2, len-2);
                     sprintf(outputFilename, "%s", filename);
                  } else {
                     ignore(argv[n]);
                  }
               } else {
                  wrong(argv[n]);
               }
            }
            break;

            case 'V': // -V
            {
               if(len == 2) {
                  if(!flags[VERBOSE]) {
                     flags[VERBOSE]=true;
                  } else {
                     ignore(argv[n]);
                  }
               } else {
                  wrong(argv[n]);
               }
            }
            break;

            case 'v': // -v
            {
               if(len == 2) {
                  if(!flags[VERSION]) {
                     flags[VERSION]=true;
                     version();

                     return 0;
                  } else {
                     ignore(argv[n]);
                  }
               } else {
                  wrong(argv[n]);
               }
            }
            break;

            default:
            {
               wrong(argv[n]);
            }
            break;
         }
      }
   }

   // --- Constants initialization ---------------------------------------------
   if(!flags[INPUT_SYSG]) { // Default input geodetic system: World Geodetic System 1984 (WGS84)
      in.sysg=0;

      in.cstes.a=6378137.0;
      in.cstes.f=1/298.257223563;

      in.cstes.e=E(in.cstes.a, in.cstes.a*(1.0-in.cstes.f)); // e=f(a, b) w/ b=a(1-f)
   }

   if(!flags[INPUT_PROJ]) { // Default input projection: -none-
      in.proj=0;
   } else if(in.sysg == 0) { // VERIF': WGS84 => no projection !
      if(in.proj != 0) {
         in.proj=0;
      }
   }

   if(!flags[INPUT_COORD]) { // Default output coordinates type: geographic
      in.coord=GEOGRAPHIQUES;
   } else { in.coord=GEOGRAPHIQUES; } // LIMIT': un seul type de coorddonnees en entree ! (geographiques)

   if(!flags[OUTPUT_SYSG]) { // Default output geodetic system: Nouvelle Triangulation Franaise (NTF)
      out.sysg=1;

      out.cstes.a=6378249.2;
      out.cstes.b=6356515.0;

      out.cstes.e=E(out.cstes.a, out.cstes.b);
      out.cstes.lambda0=deg2rad(2.337229167); // 0 grades Paris (= 2 20' 14,025" E Greenwich)

      out.cstes.T.x=-168.0;
      out.cstes.T.y=-60.0;
      out.cstes.T.z=320.0;
      out.cstes.D=0.0;
      out.cstes.R.x=0.0;
      out.cstes.R.y=0.0;
      out.cstes.R.z=0.0;
   }

   if(!flags[OUTPUT_PROJ]) { // Default output projection: Lambert II extended (NTF)
      out.proj=3;

      out.cstes.n=0.7289686274;
      out.cstes.C=11745793.39;
      out.cstes.Xs=600000.0;
      out.cstes.Ys=8199695.768;
   }

   if(!flags[OUTPUT_COORD]) { // Default output coordinates type: planes
      out.coord=PLANES;
   } else { out.coord=PLANES; } // LIMIT': un seul type de coorddonnees en sortie ! (planes)

   // --- Files initialization -------------------------------------------------
   FILE *pInputFile=NULL, *pOutputFile=NULL;

   if(flags[OUTPUT_FILE]) {
      pOutputFile=fopen(outputFilename, "wt");
   } else {
      flags[VERBOSE]=true;
   }

   if(flags[INPUT_FILE]) {
      pInputFile=fopen(inputFilename, "rt");

      if(flags[OUTPUT_FILE] && (pOutputFile == NULL)) {
         printf("MIS-convert: [ERROR] Output file can not be appended. [%s]\n", outputFilename);
      } else {
         if(pInputFile != NULL) {
            int nb=0;

            if(flags[VERBOSE]) {
               printf("RAF98 grid loading...");
            }

            int raf98loaded=loadRAF98("raf98.gra");

            if(flags[VERBOSE]) {
               printf(" ok\n");
            }

            if(raf98loaded) {
/*
               // Obtain file size.
               fseek(pFile, 0, SEEK_END); // SEEK_END (2) End of file.
               filesize=ftell(pFile);
               rewind(pFile);
*/
               char fileLine[1024]=""; // Same as in Notepad

               if(flags[VERBOSE])
               {
                  printf("Latitude   Longitude  Depth Nord               Est               Reduced Depth\n");
               }

               while(fgets(fileLine, 1024, pInputFile)) {
                  if(fileLine[0] != '#') {
                     double lat, lon, depth, he;
                     coord_cartesiennes srcTemp, resTemp;
/*
   # Format: time lat lon depth he nbSat GDOP
   954590400 47.3925625 -0.9395926 3.52 102.355 7 5.5
*/
                     sscanf(fileLine, "%*s %lf %lf %lf %lf %*s %*s\n", &lat, &lon, &depth, &he);

                     // Read input coordinates by type
                     switch(in.coord) {
                        case CARTESIENNES:
                        {/*
                           coord_cartesiennes src;

                           srcTemp=src;
                        */}
                        break;

                        case GEOGRAPHIQUES:
                        {
                           coord_geographiques src;

                           src.phiv=deg2rad(lat);
                           src.lambda=deg2rad(lon);
                           src.he=he;

                           srcTemp=cg2cc(src, in.cstes.a, in.cstes.e);
                        }
                        break;

                        default: // PLANES
                        {/*
                           coord_planes src;

                           srcTemp=cp2cc(src, ...);
                        */}
                        break;
                     }

                     resTemp=similitude3D_7params(srcTemp, out.cstes.T, out.cstes.D, out.cstes.R);

                     // Write output coordinates by type
                     switch(out.coord) {
                        case GEOGRAPHIQUES:
                        {
                           coord_geographiques res=cc2cg(resTemp, out.cstes.a, out.cstes.e);

                           if(flags[VERBOSE])
                           {
                              printf("%.7lf %.7lf %-5.2f %.10lf %.10lf %.2lf\n", lat, lon, depth, res.phiv, res.lambda, res.he);
                           }

                           if(flags[OUTPUT_FILE])
                           {
                              fprintf(pOutputFile, "%.10lf %.10lf %.2lf\n", res.phiv, res.lambda, res.he);
                           }
                        }
                        break;

                        case CARTESIENNES:
                        {
                           if(flags[VERBOSE])
                           {
                              printf("%.7lf %.7lf %-5.2f %.10lf %.10lf %.2lf\n", lat, lon, depth, resTemp.X, resTemp.Y, resTemp.Z);
                           }

                           if(flags[OUTPUT_FILE])
                           {
                              fprintf(pOutputFile, "%.10lf %.10lf %.2lf\n", resTemp.X, resTemp.Y, resTemp.Z);
                           }
                        }
                        break;

                        default: // PLANES
                        {
                           coord_planes res=cg2cp(cc2cg(resTemp, out.cstes.a, out.cstes.e), out.cstes.C, out.cstes.e, out.cstes.lambda0, out.cstes.n, out.cstes.Xs, out.cstes.Ys);

                           double alt=altitude(lat, lon, depth, he, ha);

                           if(flags[VERBOSE])
                           {
                              printf("%-10.7lf %-10.7lf %-5.2f %-18.10lf %-17.10lf %-10.7lf\n", lat, lon, depth, res.N, res.E, alt);
                           }

                           if(flags[OUTPUT_FILE])
                           {
                              fprintf(pOutputFile, "%.10lf %.10lf %.10lf\n", res.E, res.N, alt);
                           }
                        }
                        break;
                     }

                     nb++;
                  } // else: line is a comment.
               }
            } else {
               printf("MIS-convert: [ERROR] Loading `raf98.gra' failed.\n");
            }

            fclose(pInputFile);

            printf("Finished.\n");

            if(flags[OUTPUT_FILE]) {
               printf("Output file: %s [%d line%s\n", outputFilename, nb, ((nb > 1) ? "s]" : "]"));
            }
         } else {
            printf("File not found. [%s]\n", inputFilename);
         }
      } // else: No output-file specified.
   } else {
      printf("MIS-convert: [ERROR] Need <input-file> specidied.\n");
   }

   if(flags[OUTPUT_FILE]) {
      fclose(pOutputFile);
   }

   return 0;
}
